#!/usr/local/bin/perl
#
# Copyright 2017, Jim Brown.
# This work is released under the BSD 2-clause license, for educational
# purposes only.  The author disclaims any responsibility for the use
# of this tool.  Enjoy!
#
# Version 2.0 - Mon Apr 10 12:38:50 CDT 2017
#

use strict;


# short help
sub usage() {

  print << "EOT";

perl comboleetor.pl  [-b blocksfile] [-n numbersfile] [-c] [-s] [-h | -H]

     where blocksfile and numbersfile contain word blocks and numbers respectively,
     default files are blocks.txt and numbers.txt in the current directory.

     -c  Serialize capitalization throughout the output word
     -s  Print all hashes and exit
     -h  Print short help (this text)
     -H  Print longer help text with examples

EOT
  exit;
}



# longer help
sub usage2() {

  print << "EOT2";

perl comboleetor.pl  [-b blocksfile] [-n numbersfile] [-c] [-s] [-h | -H]

     where blocksfile and numbersfile contain word blocks and numbers respectively,
     default files are blocks.txt and numbers.txt in the current directory.

     -c  Serialize capitalization throughout the output word
     -s  Print all hashes and exit. Allows you to see the contents
         of all hashes just prior to combining.  Useful for debugging.
     -h  Print short help
     -H  Print longer help text with examples (this text)

comboleetor.pl - password generator combining
                 blocks, numbers, punctuation,
                 and 'leetspeak' substitutions.

comboleetor.pl combines word blocks, numbers, and punctuation in any
order determined by an input specification.  The resulting output can be
further subjected to 'leetspeak' substitutions and serial capitalization.

The input specification allows for up to five characters of the following
set in any order:

   B  word blocks

   b  word blocks subjected to leet substitutions at output.

   N  Number elements

   P  Punctuation elements


The goal is to produce all possible combinations of these elments
in the order they are specified.  The specification of elements
comes from the user on stdin via terminal, pipe, file, etc.

For example, consider the blocks file (default 'blocks.txt')
containing two words:
   tin
   can

and the numbers file (default 'numbers.txt') containing three
numbers
   031
   072
   2600

Running comboleetor.pl like this

   echo 'BN' | perl comboleetor.pl

produces the following on stdout:

   can
   can2600
   can031
   can072
   tin
   tin2600
   tin031
   tin072

Another example:

   echo 'NBN' | perl comboleetor.pl

produces on stdout:

   can
   can2600
   can031
   can072
   tin
   tin2600
   tin031
   tin072

Another example:

   echo 'NBN' | perl comboleetor.pl

produces on stdout:

   2600
   2600can
   2600can2600
   2600can031
   2600can072
   2600tin
   2600tin2600
   2600tin031
   2600tin072
   031
   031can
   031can2600
   031can031
   031can072
   031tin
   031tin2600
   031tin031
   031tin072
   072
   072can
   072can2600
   072can031
   072can072
   072tin
   072tin2600
   072tin031
   072tin072


The -c options provides serialized capitalization.  Each letter
is capitalized in turn.

  echo 'B' | perl comboleetor.pl -c

produces on stdout:

   can
   can
   Can
   cAn
   CAn
   caN
   CaN
   cAN
   CAN
   tin
   tin
   Tin
   tIn
   TIn
   tiN
   TiN
   tIN
   TIN


With 'b', there is leetspeak substitution. The next example
shows interactive use of the program from the terminal.
Some additional output to stderr is shown. To output just the
passwords, redirect stdout to a separate file.

   \$ perl comboleetor.pl
   :>> b
   BLOCKS with leet subs.


   ===========================================================


   can
   can
   (an
   c4n
   (4n
   ca|\\|
   (a|\\|
   c4|\\|
   (4|\\|
   can
   (an
   c\@n
   (\@n
   ca|\\|
   (a|\\|
   c\@|\\|
   (\@|\\|
   can
   (an
   c/-\\n
   (/-\\n
   ca|\\|
   (a|\\|
   c/-\\|\\|
   (/-\\|\\|
   tin
   tin
   7in
   t!n
   7!n
   ti|\\|
   7i|\\|
   t!|\\|
   7!|\\|
   tin
   +in
   t!n
   +!n
   ti|\\|
   +i|\\|
   t!|\\|
   +!|\\|
   :>> ^D
   END

   \$

If you do not see your favorite leet substitution, edit the code and
re-run.

This can also be combined with the -c option, which capitalizes
any lower case characters that appear in the leet output:

   \$ perl comboleetor.pl -c
   :>> b
   BLOCKS with leet subs.

   ===========================================================

   (4|\\|
   +i|\\|
   +i|\\|
   +I|\\|
   c/-\\|\\|
   c/-\\|\\|
   C/-\\|\\|
   ca|\\|
   ca|\\|
   Ca|\\|
   cA|\\|
   CA|\\|
   can
   can
   Can
   cAn
   CAn
   caN
   CaN
   cAN
   CAN
   (\@|\\|
   ti|\\|
   ti|\\|
   Ti|\\|
   tI|\\|
   TI|\\|
   +in
   +in
   +In
   +iN
   +IN
   (\@n
   (\@n
   (\@N
   c\@|\\|
   c\@|\\|
   C\@|\\|
   c/-\\n
   c/-\\n
   C/-\\n
   c/-\\N
   C/-\\N
   t!n
   t!n
   T!n
   t!N
   T!N
   (a|\\|
   (a|\\|
   (A|\\|
   c4n
   c4n
   C4n
   c4N
   C4N
   t!|\\|
   t!|\\|
   T!|\\|
   c\@n
   c\@n
   C\@n
   c\@N
   C\@N
   (/-\\|\\|
   c4|\\|
   c4|\\|
   C4|\\|
   +!|\\|
   (an
   (an
   (An
   (aN
   (AN
   tin
   tin
   Tin
   tIn
   TIn
   tiN
   TiN
   tIN
   TIN
   (4n
   (4n
   (4N
   (/-\\n
   (/-\\n
   (/-\\N
   +!n
   +!n
   +!N
   7in
   7in
   7In
   7iN
   7IN
   7i|\\|
   7i|\\|
   7I|\\|
   7!|\\|
   7!n
   7!n
   7!N
   :>> ^D
   END
   \$

Note that capitalization with -c happens last, just before output.


Try the following combinations interactively:

   BBN
   BNb
   bNPP
  
etc.

Note: this program can generate a *lot* of output.  Start with small files
and simple patterns to ensure you know how the program operates.

Also, output is not guaranteed to be unique.  Certain combinations of
options may generate identical output, though the amount of non-unique
output is relatively small.  To produce unique output, postprocess
the output with other tools such as:

  echo 'bb' | perl comboleetor.pl -c | sort | uniq


NUMBERS and NUMBER FORMATS

Starting in version 2.0, the usage of the numbers file has changed.
Numbers can still be specified one line at a time, but a new syntax
is available to generate lists and ranges of numbers with an optional
format specification.


Lines may now take the form:

   n | n,n... | n-m | printspec | # Comment...

 where printspec is of the form:

   %[-|+][n][.n][d|x]

   Examples:

     %-03.3d   - left justified, zero filled 3 octet decimal number field

     %-05.5x   - left justified, zero filled 5 octet hexidecimal number field

     %d        - any decimal number, no format applied

The printspec specification is based on the sprintf format definitions.
See the "perldoc -f sprintf" function documentation for format details.
Generally speaking, in practice, only the 'd' and 'x' format specifications
are of any use.  The range syntax makes it easy to create a large number
of entries, e.g. 0-9999 produces 1000 entries in the numbers hash.

Format specifications remain in effect until changed, and numbers and or
ranges may be overridden by later entries.

For example, consider a numbers file containing:

#
#  Numbers file demonstrating v2.0 syntax
#
#
# Start with a print spec of a 3 character zero filled decimal numbers
#
%03.3d
31,72,2600
#
#
# Switch to 5 char zero filled hexadecimal and output a range
#
%05.5x
40-50
#
#
# Remove zero fills and switch back to decimal list and range
#
%d 
1,2,4,10-15


produces a numbers hash containing:


1
2
4
10
11
12
13
14
15
031
00028
00029
0002a
0002b
0002c
0002d
0002e
0002f
00030
00031
00032
072
2600



#
# Note: simple is good.  Try to keep the numbers and formats simple.
# Use the -s option to output all hashes to examine the number entries.
#



<Author Rant>
Ok, why another combinator?

I needed to recover a password I used over 10 years ago.  Back then, I used
to use personal word blocks, numbers, and punctuation in different
uh, combinations.  Since the end results were usually 10 or more characters
or (frequently 12-15) just brute forcing it was not practical.  Also, my
use of leetspeak is a bit different than most, as I sometimes used multiple
characters in my leet substitutions - '|_' for 'L' and other similar
subs as you can see in the code.

This combinator drastically reduces the search time by only using blocks
of text, numbers, punctuation, and leetspeak frequently used.

It is missing some features found in other combinators - rotation, space
generation, etc.  These can be easily added, or the output can be piped to
another combinator that does those operations.

If someone else has already written something just like it, I've yet
to find it though, to be honest, I didn't look really hard :-).

I hope it's useful to someone else besides me.

If you like it, drop me a line - jpb Atsign jimby.name
</Author Rant>


EOT2

  exit;
}






use Getopt::Std;

#
# Option stuff
#

my %cmdopts = ();
# [-b blockfile] [-n numbersfile] [-l] [-h | -H]
getopts("b:n:clhHs", \%cmdopts);


# Start with no capitilization substitutions unless explicitly requested
our $cap_flag = ($cmdopts{c} ? 1 : 0);


#
# Do usage subs first if requested
#


# short help and exit
if($cmdopts{h}) { usage(); }

# long help and exit
if($cmdopts{H}) { usage2(); }



my $blockfile   = ($cmdopts{b} ? $cmdopts{b} : "blocks.txt");
my $numbersfile = ($cmdopts{n} ? $cmdopts{n} : "numbers.txt");
#


#
# Part I.  Get the hashes  and variables we need all set up.
#


my %blocks = ();
my %nums   = ();

my $output = "";

open(BLOCKFILE, "<", $blockfile)   or die "Can't open blockfile [$blockfile] : $!\n";
open(NUMFILE,   "<", $numbersfile) or die "Can't open numbers file [$numbersfile] : $!\n";


# Slurp in text blocks.
while(<BLOCKFILE>)
{
   my  $inputline = $_;
   chop $inputline;

   next if ($inputline =~ /^\s*#/);    # Skip comments

   $blocks{$inputline} = $inputline;
#  print "blocks{$inputline}\n";

}



# printspec is for formatting numbers with the v2.0 number code
my $printspec = "";

# Slurp in numbers.
while(<NUMFILE>)
{
   my  $inputline = $_;
   chop $inputline;

   next if ($inputline =~ /^\s*#/);    # Skip comments

   &numhash($inputline);

#   $nums{$inputline} = $inputline;
#  print "nums{$inputline}\n";
}


#
# Lots of leet subs taken from http://i.imgur.com/vImVwZV.png
# which I think is from
# https://lingojam.com/L33tTranslator(Advanced)
# Surely not complete...
#
# Some letters have more than one leet variant.
# Note that the maximum number of variations is 3.
# We use this fact to create three separate hashes,
# a primary hash, and two alternates.
#
# Consider these three variations on the base word 'can'
# on the second character of the word - the letter 'a'
#
#   c4n
#   c@n
#   c/-\n
#
# These are all variants of the leet substituion for 'a'.
#
# To generate all three variants, we use the primary hash
# first and generate the word. Then we copy over the
# hash values from the first alternate hash to the primary
# hash (replacing the hash value for '4' with '@' for
# example) and generate the word again with the new values.
# We then copy over the hash values for the second
# alternate to the primary (replacing the hash value
# for '@' again with '/-\') and generate the word a third time.
#
# Note that to optimize all this, we want to examine
# the incoming text to see if any alt_1 and  alt_2 letters
# are in the target word. If not, there is no need to copy
# and regenerate. The target word 'grip' is one such example.
# There are no alt_1 or alt_2 letters in this word.
#
# Also note, we keep the %leetz_orig hash as the backup copy
# to restore the primary hash for the next word.
#
#
#

# Hardcoded leetz hash (standard set and backup)

my %leetz_orig = (
  'a'   => '4',
  'A'   => '4',
  'b'   => '|o',
  'B'   => '8',
  'c'   => '(',
  'C'   => '(',
  'd'   => '|)',
  'D'   => '|)',
  'e'   => '3',
  'E'   => '3',
  'f'   => '|=',
  'F'   => ']=',
  'g'   => '9',
  'G'   => '6',
  'h'   => '#',
  'H'   => '|-|',
  'i'   => '!',
  'I'   => '|',
  'j'   => '_|',
  'J'   => '_|',
  'k'   => 'X',
  'K'   => '|<',
  'l'   => '1',
  'L'   => '|_',
  'm'   => '^^',
  'M'   => '|\\/|',
  'n'   => '|\\|',
  'N'   => '|\\|',
  'o'   => '0',
  'O'   => '0',
  'p'   => '|*',
  'P'   => '|*',
  'q'   => '9',
  'Q'   => '0_',
  'r'   => '|^',
  'R'   => '/2',
  's'   => '5',
  'S'   => '5',
  't'   => '7',
  'T'   => '7',
  'u'   => '|_|',
  'U'   => '|_|',
  'v'   => '\\/',
  'V'   => '\\/',
  'w'   => 'vv',
  'W'   => '\\/\\/',
  'x'   => '%',
  'X'   => '><',
  'y'   => 'j',
  'Y'   => '`/',
  'z'   => '2',
  'Z'   => '2',
  );

my %leetz = %leetz_orig;



# First alternate set, aka alt_1 in code and comments.
my %leetz_subs1 = (
  'a'   => '@',
  'A'   => '@',
  'B'   => '|3',
  'E'   => '[-',
  'H'   => ']-[',
  'l'   => '|',
  'M'   => '/\\/\\',
  'N'   => '/\\/',
  'O'   => '()',
  's'   => '$',
  'S'   => '$',
  't'   => '+',
  'T'   => '+',
  'u'   => '(_)',
  'U'   => '(_)',
  'w'   => '\\|/',
  'W'   => '\\_|_/',
  'x'   => '><',
  'y'   => '`/',
  'z'   => '%',
  'Z'   => '7_',
  );





# Second alternate set, aka alt_2 in code and comments.
my %leetz_subs2 = (
  'a'   => '/-\\',
  'A'   => '/-\\',
  'U'   => '\\_/',
  );



# Punctuation included:  !"#$%&'()*+,-./:;<=>?@[\]^_`{|}~

# Key names in this hash are sorted later in the code, thus
# the number prefix.

my %punct = (
  '00sp'  => ' ',
  '01ex'  => '!',
  '02dq'  => '"',
  '03ha'  => '#',
  '04do'  => '$',
  '05pc'  => '%',
  '06am'  => '&',
  '07sq'  => '\'',
  '08lp'  => '(',
  '09rp'  => ')',
  '10fu'  => '*',
  '11pl'  => '+',
  '12cm'  => ',',
  '13da'  => '-',
  '14pd'  => '.',
  '15fs'  => '/',
  '16co'  => ':',
  '17sc'  => ';',
  '19la'  => '<',
  '19eq'  => '=',
  '20ra'  => '>',
  '21qu'  => '?',
  '22at'  => '@',
  '23rb'  => '[',
  '24bs'  => '\\',
  '25lb'  => ']',
  '26ca'  => '^',
  '27un'  => '_',
  '28bt'  => '`',
  '29lc'  => '{',
  '30pi'  => '|',
  '31rc'  => '}',
  '32ti'  => '~',
  );



#
# %leethash keeps a complete hash of all leet substitutions for all
# elements in the %blocks hash. We do this to optimize the way we
# combine blocks and leetz  i.e input of   bB  
# Filled once before all processing.  See the -s option to 
# show the filled in hash.
#

my %leethash = ();

my $k = "";
my $v = "";
our $print_leethash = 0; #turn off printing leethash in leetorator()
our $print_caps     = 0; #turn off printing caps in leetorator()

foreach $k (sort keys (%blocks))
{
  %leetz = %leetz_orig;  # make sure it's the original leetz hash
  $v = $blocks{$k};
  #print "blocks{$k} = [$v]\n";
  &leetorator($v);   # sets this word in %leethash
}


$print_leethash = 1; #turn on printing leethash
$print_caps     = 1; #turn on printing caps



# We have populated all the hashes at this point.
# If -s was requested, show hashes and exit
if($cmdopts{s}) { &prhashes(); exit; }



#
# We have a 5 level combinator engine.  These are the hashes we will use.
#

my %level0_hash = ();
my %level1_hash = ();
my %level2_hash = ();
my %level3_hash = ();
my %level4_hash = ();


# This array is used to resort hashes before the combinator engine starts.
my @AoH = ();



# Some needed text variables for each leveln_hash.
my $text0 = "";
my $text1 = "";
my $text2 = "";
my $text3 = "";
my $text4 = "";


#
# Part II.  Get and decode input specification.
#


print STDERR ':>> ';

#
# Read and decode a specification on stdin.
#


# Main loop - handles input.
while(<>)
{

  %leetz = %leetz_orig;      # Start with a clean leetz hash
  my $inputline = "";
  my $inputf = "";
  my @inputfields = ();



  $inputline = $_;
  chop $inputline;


  my $bref  = \%blocks;       # Each scalar holds a reference to the listed array
  my $nref  = \%nums;
  my $lref  = \%leetz;
  my $pref  = \%punct;
  my $lhref = \%leethash;

  my $AoH_index = 0;

  @inputfields = split(//,$inputline);   # Break up the input specification so we can apply each combination.


# The combinator engine below depends on hashes in known order- level0, level1, level2,  etc.
# Input from the user comes in an unknown order so we need to map the hashes from the input order
# to the known order. To do this we place  input order into an array of hashes AoH.
# The code in Part III will decode the mappings in the order requested into the known hash order
# (level0_hash, level1_hash, etc.)

  while($inputf = shift @inputfields)
  {
    # Put references into the AoH array if a valid specification char was read.  Bump AoH_index if so.
    if ($inputf =~/^B$/)
        { $AoH[$AoH_index] =  $bref; $AoH_index++; print STDERR "BLOCKS\n";}

    if ($inputf =~/^b$/)
        { $AoH[$AoH_index] =  $lhref; $AoH_index++; print STDERR "BLOCKS with leet subs.\n";}

    if ($inputf =~/^N$/)
        { $AoH[$AoH_index] =  $nref; $AoH_index++; print STDERR "NUMS\n";}

    if ($inputf =~/^P$/)
        { $AoH[$AoH_index] =  $pref; $AoH_index++; print STDERR "PUNCT\n";}

    if ($inputf =~/^\.$/)  # undocumented  - print all hashes while running
        { &prhashes();}

    if ($inputf =~/^c$/)   # undocumented  - toggle capitalization
        { $cap_flag ^= 1;}
  }

  
  
  print STDERR "\n\n===========================================================\n\n";
  
#
# Part III.  Decode input and get ready to combo.
#
  
  
  # testing @AoH = ( [ %blocks ] , [ %blocks ], [ %punct ], [ %nums ] );

  # Remapping.  Grab the references from the AoH array and remap into the
  # %leveln_hashes.
  #
  # AoH_index is pointing after the last filled slot, so
  # we have to subtract 1 for the logic below.
  
  if (($AoH_index - 1) >= 0)
  {
    %level0_hash =  %{ $AoH[0] };
  }
  else
  {
    %level0_hash = ();
  }


  if (($AoH_index - 1) >= 1)
  {
    %level1_hash =  %{ $AoH[1] };
  }
  else
  {
    %level1_hash = ();
  }


  if (($AoH_index - 1) >= 2)
  {
    %level2_hash =  %{ $AoH[2] };
  }
  else
  {
    %level2_hash = ();
  }

  
  if (($AoH_index - 1) >= 3)
  {
    %level3_hash =  %{ $AoH[3] };
  }
  else
  {
    %level3_hash = ();
  }

  
  if (($AoH_index - 1) >= 4)
  {
    %level4_hash =  %{ $AoH[4] };
  }
  else
  {
    %level4_hash = ();
  }

  
  # Five level combinator - should be enough for most.
  # Too lazy to use recursion. 

  # Let 'er rip!
  
  $output = "";
  
  for  my $level0_key (sort {$a <=> $b} keys %level0_hash) {
    $text0 = $level0_hash{$level0_key};
    #print "0:[$level0_key][$text0]\n";
    $output = $text0;
    print "$output\n";
    if($cap_flag  == 1)
    { 
      postprocess($output)
    };
  
  
  
    for my  $level1_key (sort {$a <=> $b} keys %level1_hash) {
      $text1 = $level1_hash{$level1_key};
      #print "1:[$level1_key][$text1]\n";
      $output .= $text1;
      $output = $text0 . $text1;
      print "$output\n";
      if($cap_flag  == 1)
      {
         postprocess($output)
      };
     
  
      for my  $level2_key (sort {$a <=> $b} keys %level2_hash) {
        $text2 = $level2_hash{$level2_key};
        #print "2:[$level2_key][$text2]\n";
        $output .= $text2;
        $output = $text0 . $text1 . $text2;
        print "$output\n";
        if($cap_flag  == 1)
        {
          postprocess($output)
        };
  
  
  
        for my  $level3_key (sort {$a <=> $b} keys %level3_hash) {
          $text3 = $level3_hash{$level3_key};
          #print "3:[$level3_key][$text3]\n";
          $output .= $text3;
          $output = $text0 . $text1 . $text2 . $text3;
          print "$output\n";
          if($cap_flag  == 1)
          {
            postprocess($output)
          };
  
  
          for my  $level4_key (sort {$a <=> $b} keys %level4_hash) {
            $text4 = $level4_hash{$level4_key};
            #print "4:[$level4_key][$text4]\n";
            $output .= $text4;
            $output = $text0 . $text1 . $text2 . $text3 . $text4;
            print "$output\n";
            if($cap_flag  == 1)
            {
              postprocess($output)
            };
  
          }

        }
      }
    }
  }
  

  print STDERR ':>> ';
  
}  # End main loop.
  
  
print STDERR  "\nEND\n\n";

exit;



sub postprocess() {
  (my $targetword) = @_;

  %leetz = %leetz_orig;   # Start with original leetz hash
  # We process capitilization inside the emitter sub (right before emitting the output word).
  if ($cap_flag   == 1) { alpharator($targetword); };

}



sub leetorator() {
 (my $myword) = @_;

  my $keep_going = 0;
  my $i = "";

  # Generate leet subs using the standard hash first.
  $keep_going = &emitter($myword);

  if($keep_going > 0)
  {
    # A character in the alt_1 hash was found in the target word.
    # Substitute the alt_1 letters and re emit.

    foreach  $i (sort keys %leetz_subs1)
    {
       #print "[$i] [$leetz{$i}]  ";
       $leetz{$i} = $leetz_subs1{$i};   # replace with new leet value
       # print "[$leetz{$i}]\n";
    }

    $keep_going = &emitter($myword);    # generate with alt_1 subs

    # We have already done alt_1 characters.  Only
    # alt_2 characters are left.  Alt_2 is a proper subset
    # of alt_1, so this nested-if construction should be fine.


    if($keep_going  == 2) 
    {
      # Substitute the alt_2 letters and re emit.
      foreach $i (sort keys %leetz_subs2)
      {
         #print "[$i] [$leetz{$i}]  ";
         $leetz{$i} = $leetz_subs2{$i};
         #print "[$leetz{$i}]\n";
      }

      # return value is not used now
      my $junk = &emitter($myword);

    } # end inner if
  } # end outer if

}



 
# emitter, the leet word generation routine.
sub emitter() {
   (my $targetword) = @_;

  #  
  # emitter
  #
  # - Process for leet substitution
  # - Using a bitstring, come up with all permutations
  #   of $word with leet replacements for each characters.
  #   The bitstring is set each time through the outer loop, and
  #   becomes a binary sequence from left to right.
  # 
  # Example:
  #
  # substitution  bitstring         leetz_hash{$t} replacement
  #-------------------------------------------------------------------------
  #   abedi       0000000000000...  didn't replace anyting
  #   4bedi       1000000000000...  replaced first character
  #   a|oedi      0100000000000...  second character by two characters '|o'
  #   4|oedi      1100000000000...  replaced first and second characters
  #   ab3di       0010000000000...  replaced third character only
  #   4b3di       1010000000000...  replaced first and third
  #   a|o3di      0110000000000...  etc., continues in binary sequence
  #   4|o3di      1110000000000...  
  #   abe|)i      0001000000000... 
  #   4be|)i      1001000000000... 
  #   etc...


  my $i = 0;
  my $j = 0;

  my $wordlength = length($targetword);
  my @wordbytes  = split(//,$targetword);
  my @leetchars = ();
  my $outword = "";
  my $bitsindex = 0;

  if($wordlength == 0) { return 0; } # If wordlength is 0, there is nothing to do.
  
  for ($i = 0; $i < 2**$wordlength;  $i++)
  {
    # Use a bit string array to keep track of what to do in each pass 
    my $vector = pack("Q", $i);            # 64 bit maximum - should be enough for most uses
    my $bits = unpack("b64", $vector);     # Bit string for current outerloopcount
    # print "i: $i: bits: [$bits]\n";
    my @bits = split(//, unpack("b64", $vector));  # bit array for substitutions for this round
     
  
    $outword = "";
    $bitsindex = 0;

    # The inner loop iterates over all bytes in the target word  (@wordbytes)
    # As we go though the loop, if we find a leet character, and if we are supposed to
    # substitute it in this round of the outerloop, we substitute it, or not.
    # We then increment $bitsindex.  It is the $bits[$bitsindex] variable we need to
    # examine to see if we are supposed to substitute the current candidate.
    # In either case, we copy chars to $outword as we go, and emit the result at the bottom
    # of each loop iteration.
  
    for($j = 0; $j < $wordlength; $j++)
    {
      my $t = $wordbytes[$j];
  
      if($bits[$bitsindex] == '1')
      {
        if($leetz{$t} ne "")        # only sub if we have a valid char in %leetz
        {
          @leetchars = split /,/,$leetz{$t};
          my $aleet = pop @leetchars;  # just one for now
          $outword .= $aleet;
        }
        else
        {
          $outword .= $t;
        }
      }
      else
      {
        $outword .= $t;
      }
      $bitsindex++;
    }
  
    #print "$outword\t\t$bits\n" if ($main::print_leethash == 1);  # for debugging
    print "$outword\n"           if ($main::print_leethash == 1); # defined in main

    $leethash{$outword} = $outword;

    # We also process capitilization inside the leetorator sub (right after emitting the output word)
    if ($cap_flag   == 1 &&
        $main::print_caps == 1) { alpharator($outword); };
  }  


  # Here is where we check to see if any alt_2 and alt_1
  # letters are found in the word.  If alt_1 is found return 1
  # if alt_2 is found return 2 otherwise, return 0

  my $retval = 0;                      # initialize 
  my $alt2_retval = 0;
  my $alt1_retval = 0;
  my $check_alt = "";

  foreach $check_alt (@wordbytes)
  {
    if($check_alt =~ /[aAU]/)       # check for alt_2 chars first
    {
      # print "check_alt_2 [$check_alt] found\n";
      $alt2_retval = 2;  # set return value
    }
    # if($leetz_subs1{$check_alt} ne "")  # now check for alt_1
    if($check_alt =~ /[BEHlMNOsStTuwWxyzZ]/)
    {
      # print "check_alt [$check_alt] found\n";
      $alt1_retval = 1;  # set return value
    }
  }

  # There is an implied priority here.  Return 2 if the alt_2
  # check was successful, 1 if alt_1 was successful, or
  # zero if neither was succeessful.
  # The priority is utilized by the caller of this function.

  if($alt2_retval == 2)
  {
    # print "retval = $alt2_retval\n";
    return $alt2_retval;
   }

  if($alt1_retval == 1)
  {
    # print "retval = $alt1_retval\n";
    return $alt1_retval;
   }

  return $retval;   # zero
}

 
 
 
#
# countalpha()
#
#   Count number of all chars in the target word
#   and return the count, and an array of chars found
#

sub countalpha() {
   (my $inword) = @_;
   my @allchars = split //, $inword;

   my $alphacount = 0;
   my @alpharr = ();

  for (my $i = 0; $i <= $#allchars; $i++)         # Examine each char 
  {
    my  $ch = $allchars[$i];
    #print "[$ch]\n";

    if( $ch =~ /[a-z]/  )                    # Is it a lower case  char?
    {
      # $allchars[$i] = uc $ch;
      # $outword = join ('',@allchars);
      # print "out:[$outword]\n";
      $alphacount++;                            # Bump count
      push @alpharr, $ch;
    }
  }

  return ($alphacount, @alpharr);               # Return count and array
                                               # Array is not used at this time
}


 
sub alpharator() {
   (my $targetword) = @_;

  #  
  # alpharator
  #
  # - Serially alphabetize all letters in the string,
  #   starting on the left.  This is similar code but not
  #   identical code to leetorator() just similifed for
  #   ASCII chars that are class [:lower:]
  # 
  # Example:
  #    
  #  modification    bitstring
  # -----------------------------------------
  #    abe           000000...
  #    Abe           100000...
  #    aBe           010000...
  #    ABe           110000...
  #    abE           001000...
  #    AbE           101000...
  #    aBE           011000...
  #    ABE           111000...
  #
 
  my  $word = $targetword;
   
  # test strings
  #$word = "programmer";
  #$word = "aeioulmnaeioulmnaeioulmnaeioulmnaeioulmnaeioulmnaeioulmnaeioulmn";
  #$word = "abedi";
  #
  
  
  # Gather some essential information on the target word 
  
   my @wordbytes = split(//,$word);
  
   my $wordlength = length($word);
  
  # alpharray not really used yet.
  (my $numalpha, my @alpharray) = &countalpha($word);

  if ($numalpha == 0) { return; }     # if no alpha chars, nothing to do
  
   my $outerloopcount = 2**$numalpha; # Number of possible permutations for alphabetic substitutions
  
  for (my  $i = 0 ; $i < $outerloopcount; $i++ ) # Less than, not less than or equal
  {
     my $outword = "";                      # Initialize the output word
  
     my $vector = pack("Q", $i);            # 64 bit maximum - should be enough for most uses
     my $bits = unpack("b64", $vector);     # Bit string for current outerloopcount
    # print "i: $i: bits: [$bits]\n";
     my @bits = split(//, unpack("b64", $vector));  # bit array for substitutions
  
         
    # The inner loop iterates over all bytes in the target word  (@wordbytes)
    # As we go though the loop, if we find a lowercase character, and if we are supposed to
    # substitute it in this round of the outerloop, we substitute it, or not.
    # We then increment $bitsindex.  It is the $bits[$bitsindex] variable we need to
    # examine to see if we are supposed to substitute the current candidate.
    # In either case, we copy chars to $outword as we go, and emit the result at the bottom
    # of each loop iteration.
  
    my $bitsindex = 0;                     # Reset for each pass through outer loop

    for (my $j = 0; $j < $wordlength; $j++) # Loop through the target word
    {
      my  $t = $wordbytes[$j];
      #if($leetz{$t} ne "")               # Is this a candidate for subbing?
      if($t =~ /[a-z]/)               # Is this a candidate for subbing?
      {
        if ($bits[$bitsindex] == '1')    # If 1, we do the substitution
        {
           #$sub = $leetz{$t};            # Substitute
          my  $sub = uc ($t);                # Substitute
          $outword .= $sub;              # And copy to this output
        }
        else
        {
          $outword .= $t;
        }
        $bitsindex++;                    # Bump pointer into the bits array, to look at next candidate
      }
      else
      {
        $outword .= $t;                  # We did not hit a leetz character, just output the character
      }

    } # End inner loop
 
  #
  # Emit current word.
    print "$outword\n";
    #print "$outword\t\t$bits\n";  # debug
  
  }  # End outter loop

} # End sub




sub prhashes() {
  (my $junk) = @_;

  my $k = "";

  print "\nBlocks Hash\n";
  print "-------------------------------------\n";

  foreach $k (sort keys(%blocks))
  {
    print "blocks{$k} :  $blocks{$k}\n";
  }
  
  
  
  print "\nNumbers Hash\n";
  print "-------------------------------------\n";
  
  foreach $k (sort {$a <=> $b} keys(%nums))
  {
    print "nums{$k} :  $nums{$k}\n";
  }
  
  
  
  print "\nMain Leetz Hash\n";
  print "-------------------------------------\n";
  
  foreach $k (sort keys(%leetz))
  {
    print "leetz{$k} :  $leetz{$k}\n";
  }
  
  
  
  print "\nAlternate 1 Leetz Hash\n";
  print "-------------------------------------\n";
  
  foreach $k (sort keys(%leetz_subs1))
  {
    print "leetz_subs1{$k} :  $leetz_subs1{$k}\n";
  }


  
  print "\nAlternate 2 Leetz Hash\n";
  print "-------------------------------------\n";
  
  foreach $k (sort keys(%leetz_subs2))
  {
    print "leetz_subs2{$k} :  $leetz_subs2{$k}\n";
  }

  
  
  print "\nPunctuation Hash\n";
  print "-------------------------------------\n";
   
  foreach $k (sort keys(%punct))
  {
    print "punct{$k} :  $punct{$k}\n";
  }
 
  
  print "\nLeethash Hash\n";
  print "-------------------------------------\n";
   
  foreach $k (sort keys(%leethash))
  {
    print "leethash{$k} :  $leethash{$k}\n";
  }
 
  # Do not exit.  
 
}



# numhash - 
#
# Read numbers and fill hash according to a specification.
# The number file should have lines of the form
#
#   n | n,n... | n-m | printspec
#
# where printspec is of the form:
#
#   #%[-|+][n][.n]d
#
#   Examples:
#
#     #%-05.5d   - left justified, zero filled 5 char number field
#
#     #%d        - any number, no format applied
#
#  See the perldoc -f sprintf function for format details.
#


sub numhash () {
  (my $myline) = @_;


  #check if a valid specification line exists
  #like the following line:
  #
  #   %-05.5d
  #
  # If this line exists, take it as a print specification
  # for later output.  See below.  Note that it's anchored
  # to the beginning of the line, so if you want to comment
  # it out, just use "# " before it.

  # print STDERR "myline = [$myline]\n";

  my $outnum = "";

  if($myline  =~ /^\s*(\%)(\S+)\s*$/)
    {
      $printspec = $1 . $2;
      #  print STDERR "printspec[$printspec]\n";
      return;
    }

  my @nspecs = split/,/,$myline;

  foreach my $spec (@nspecs)
  {
    #print "spec: [$spec]\n";

    if($spec =~ /\s*(\d+)\-(\d+)\s*$/)
    {
      my $s1 = $1;
      my $s2 = $2;

      for (my $i = $s1; $i <= $2; $i++)
      {
        $outnum = sprintf($printspec,$i);
        $nums{$i} = $outnum;
        #print "$i\n";
      }

    }
    elsif ($spec =~/\s*(\d+)\s*$/)
    {
      my $s1 = $1;
      $outnum = sprintf($printspec,$s1);
      $nums{$s1} = $outnum;
      #print "$s1\n";
    }
    else
    {
      print STDERR  "##Malformed number spec [$spec]\n";
    }

  }

}



